/*
pix_ps3eye.cpp
flext-based gem object for creating image source from Sony Ps3Eye cam
Copyright (c) 2010 Jean-Yves Gratius (jyg@gumo.fr)
http://jy.gratius.free.fr

---------------------------------------------------------------------
Using Code Laboratories Ps3Eye driver / SDK.
    http://codelaboratories.com/

------------------------------------------------------------------
Derived from flext tutorials

FLEXT Copyright (c) 2002-2006 Thomas Grill (gr@grrrr.org)

-------------------------------------------------------------------------


*/



#include "pix_ps3eye.h"

//for future multicam support
#define NBMAX_OF_PS3EYE 16



// instantiate static tables
int pix_ps3eye::instanciated[NBMAX_OF_PS3EYE];
bool pix_ps3eye::camera_started[NBMAX_OF_PS3EYE];
bool pix_ps3eye::enable_image_transform[NBMAX_OF_PS3EYE];
CLEyeCameraResolution pix_ps3eye::resolution[NBMAX_OF_PS3EYE];
CLEyeCameraColorMode pix_ps3eye::mode[NBMAX_OF_PS3EYE];
float pix_ps3eye::frameRate[NBMAX_OF_PS3EYE];


// instantiate pcam[] table : handles to ps3eye
CLEyeCameraInstance pix_ps3eye::pCam[NBMAX_OF_PS3EYE];

GUID pix_ps3eye::cameraGUID[NBMAX_OF_PS3EYE];

// instantiate the class
FLEXT_NEW_V("pix_ps3eye",pix_ps3eye)


//setup function : initialise static tables
/*static*/ void pix_ps3eye::setup(t_classid c)
	{
        post ("****************************************");
        post ("*  pix_ps3eye by Jean-Yves Gratius     *");
        post ("*  (for ClEyeSDK version 1.2.0.1008)   *");
        post ("*      http://jy.gratius.free.fr       *");
        post ("****************************************");
      int i;
        for (i=0;i<NBMAX_OF_PS3EYE;i++)
        {
            instanciated[i]=0;
            resolution[i]=CLEYE_QVGA;
            frameRate[i]=30;
            mode[i]=CLEYE_COLOR_RAW;
            camera_started[i]=false;
            enable_image_transform[i]=false;
        }
    }


// class constructor
pix_ps3eye::pix_ps3eye(int argc,t_atom *argv)
{
	// structures for handling cameras when specified by GUID
    GUID Guid1;
	GUID Guid2;
	bool cameraGUID_found=true;

	// define inlets
	AddInAnything();  // add inlet of type anything (index 0)
    AddOutAnything();

	// register methods
	FLEXT_ADDMETHOD_F(0,"gem_state", start_cam);
    FLEXT_ADDMETHOD_(0,"gem_state", gem_state);

    // register method for tag "fps" and int argument
    FLEXT_ADDMETHOD_F(0,"fps",set_frameRate);

    // register method for tag "color" and variable arguments
    FLEXT_ADDMETHOD_(0,"format",set_resolution);

    FLEXT_ADDMETHOD_I(0,"gain",set_gain);

    FLEXT_ADDMETHOD_I(0,"AGC",set_AGC);

    FLEXT_ADDMETHOD_I(0,"exposure",set_exposure);
    FLEXT_ADDMETHOD_I(0,"AEC",set_AEC);
    FLEXT_ADDMETHOD_I(0,"red",set_red);
    FLEXT_ADDMETHOD_I(0,"green",set_green);
    FLEXT_ADDMETHOD_I(0,"blue",set_blue);
    FLEXT_ADDMETHOD_I(0,"AWB",set_AWB);
    FLEXT_ADDMETHOD_(0,"colorspace",set_colorspace);
    FLEXT_ADDMETHOD_I(0,"upsidedown",set_upsidedown);
    FLEXT_ADDMETHOD(0,start_cam);
    FLEXT_ADDMETHOD_(0,"reset",reset_cam);
    FLEXT_ADDMETHOD_I(0, "enable_image_process",set_image_transform_mode);
    FLEXT_ADDMETHOD_I(0, "hflip",set_hflip);
    FLEXT_ADDMETHOD_I(0, "vflip",set_vflip);
    FLEXT_ADDMETHOD_I(0, "hkeystone",set_hkeystone);
    FLEXT_ADDMETHOD_I(0, "vkeystone",set_vkeystone);
    FLEXT_ADDMETHOD_I(0, "xoffset",set_xoffset);
    FLEXT_ADDMETHOD_I(0, "yoffset",set_yoffset);
    FLEXT_ADDMETHOD_I(0, "rotation",set_rotation);
    FLEXT_ADDMETHOD_I(0, "zoom",set_zoom);
    FLEXT_ADDMETHOD_I(0, "lenscorrection1",set_lenscorrection1);
    FLEXT_ADDMETHOD_I(0, "lenscorrection2",set_lenscorrection2);
    FLEXT_ADDMETHOD_I(0, "lenscorrection3",set_lenscorrection3);
    FLEXT_ADDMETHOD_I(0, "lensbrightness",set_lensbrightness);



        // filling table of symbols with colorspace names
    color_RGBA=gensym("RGBA");
    color_RGB=gensym("RGB");
    color_BGRA=gensym("BGRA");
    color_BGR=gensym("BGR");
    color_BGR16=gensym("BGR16");
    color_Grey=gensym("Grey");
    color_GREY=gensym("GREY");
    resolution_QVGA=gensym("QVGA");
    resolution_VGA=gensym("VGA");


    // init default values
    _channels=4;

    /// obsolete
    flip=0;

    // GEM colorspace for local pix_ps3eye instance
    _colorspace=GL_RGBA;
    _resolution=resolution[_camID];

    if (_resolution==CLEYE_QVGA)
    {
        _width=320;
        _height=240;
    }
    else
    {
        _width=640;
        _height=480;
    }


    // choose which camera to open

    _camID=0;
    if(argc == 1)
        if  (IsFloat(argv[0]))  // camera specified by index
        {
            _camID=GetInt(argv[0]);
      // cameracount starts from 0
            if (_camID>0)
                _camID--;
        }


        else if (IsSymbol(argv[0])) // camera specified by GUID

        {
            char *pEnd;
            char str2[3];
            const t_symbol *s = GetSymbol(argv[0]);
            char *hex_value = s->s_name;


            if(Guid1.Data1=strtol (hex_value, &pEnd, 16 ))
                if (Guid1.Data2=-strtol (pEnd, &pEnd, 16 ))  // invert negative value because '-' is interpreted as negative sign
                    if (Guid1.Data3=-strtol (pEnd, &pEnd, 16 )) // invert negative value because '-' is interpreted as negative sign
                    {
                        pEnd+=1;                // skip '-' sign
                        for(int i=0;i<=7;i++)
                        {   strncpy (str2,pEnd,2);
                            str2[2]='\0';
                            if(str2[0]=='\0')
                                break;
                            Guid1.Data4[i]=strtol (str2, NULL, 16 );
                            pEnd+=2;
                        }
                    }
            post("Searching for camera GUID= [%08x-%04x-%04x-%02x%02x%02x%02x%02x%02x%02x%02x]...",
            Guid1.Data1, Guid1.Data2, Guid1.Data3,
            Guid1.Data4[0], Guid1.Data4[1], Guid1.Data4[2],
            Guid1.Data4[3], Guid1.Data4[4], Guid1.Data4[5],
            Guid1.Data4[6], Guid1.Data4[7]);

            cameraGUID_found=false;
            // searching for matching cam ID
            for(int j=0;j<CLEyeGetCameraCount(); j++)
            {
                Guid2=CLEyeGetCameraUUID(j);
                if ((Guid1.Data1==Guid2.Data1)&&
                    (Guid1.Data2==Guid2.Data2)&&
                    (Guid1.Data3==Guid2.Data3)&&
                    (Guid1.Data4[0]==Guid2.Data4[0])&&
                    (Guid1.Data4[1]==Guid2.Data4[1])&&
                    (Guid1.Data4[2]==Guid2.Data4[2])&&
                    (Guid1.Data4[3]==Guid2.Data4[3])&&
                    (Guid1.Data4[4]==Guid2.Data4[4])&&
                    (Guid1.Data4[5]==Guid2.Data4[5])&&
                    (Guid1.Data4[6]==Guid2.Data4[6])&&
                    (Guid1.Data4[7]==Guid2.Data4[7]))
                {
                    _camID=j;
                    cameraGUID_found=true;
                }
            }
            if (cameraGUID_found)
                post ("camera GUID found : camera No %d", _camID+1);
            else
                {
                    post("error : camera GUID not found");
                }
        }

    if (_camID>=NBMAX_OF_PS3EYE)
    {
        post("error : multicam support limited to %d cam",NBMAX_OF_PS3EYE);
        cameraGUID_found=false;
    }
    if (!cameraGUID_found)
    {
        InitProblem(); //Signal a construction problem
    }
    else
    {
    // allocate space for image buffer (choose max resolution, max channels)
    pBuffer=new BYTE[640*480*4];

    //link gem_structures

    my_pixBlock.image=my_imageStruct;

        // redefine my_pixBlock fields
    my_pixBlock.image.csize= _channels;
    if (_resolution==CLEYE_VGA)
    {
        my_pixBlock.image.xsize = _width = 640;
        my_pixBlock.image.ysize = _height = 480;
    }
    else  // QVGA resolution
    {
        my_pixBlock.image.xsize= _width= 320;
        my_pixBlock.image.ysize= _height= 240;
    }


    my_pixBlock.image.data= (unsigned char*)pBuffer;

        // data type - always GL_UNSIGNED_BYTE (except for OS X)
    my_pixBlock.image.type= GL_UNSIGNED_BYTE;

            // is this owned by us (?)
    my_pixBlock.image.notowned = 1;

    my_pixBlock.image.upsidedown = 1;//-CLEyeGetCameraParameter(pCam[_camID],CLEYE_VFLIP);
//top of image is inverted in GEM



    my_pixBlock.image.format=_colorspace;


    //Check ps3Eye and Start capture
    if(!instanciated[_camID])
    {
        // Query for PS3Eye cameras
        if (_camID>=CLEyeGetCameraCount())
        {
            // camera index too big
            post("Error : camera index %d out of total number of cams (%d)", _camID+1,CLEyeGetCameraCount());
            InitProblem(); //Signal a construction problem
        }
        else
        {
            cameraGUID[_camID]=_cameraGUID=CLEyeGetCameraUUID(_camID);
            post("camera n.%d : GUID= [%08x-%04x-%04x-%02x%02x%02x%02x%02x%02x%02x%02x]",
            _camID+1, _cameraGUID.Data1, _cameraGUID.Data2, _cameraGUID.Data3,
            _cameraGUID.Data4[0], _cameraGUID.Data4[1], _cameraGUID.Data4[2],
            _cameraGUID.Data4[3], _cameraGUID.Data4[4], _cameraGUID.Data4[5],
            _cameraGUID.Data4[6], _cameraGUID.Data4[7]);

            pCam[_camID] = CLEyeCreateCamera(_cameraGUID, mode[_camID], _resolution, 30);
            if(pCam[_camID] == NULL)
            {
                post("Error creating camera instance for cam No #d", _camID+1);
                InitProblem(); //Signal a construction problemreturn;
            }
            else
            {
                instanciated[_camID]=1;
                set_vflip(0);
            }
        //    CLEyeCameraStart(pCam[_camID]);

            // add a flag to signal that camera is not started
            camera_started[_camID]=false;

        }
    }
    else
    {
        instanciated[_camID]++;
    }
    }

}

//destructor
pix_ps3eye::~pix_ps3eye()
{

    if(instanciated[_camID])
        {
            instanciated[_camID]--;

            // check if all instances have been deleted
            if(!instanciated[_camID])
            {
                // Check if camera needs to be stopped
                if (camera_started[_camID])
                {
                    CLEyeCameraStop(pCam[_camID]);
                    camera_started[_camID]=false;
                }

                CLEyeDestroyCamera(pCam[_camID]);
                pCam[_camID]=NULL;
                }
            }
    	delete [] pBuffer;

}


void pix_ps3eye::gem_state(int argc, t_atom *argv)
{
    if (argc==2)
        {
        if (CanbePointer(argv[1]))
        {
            gem_state_pointer =  (GemState *) GetPointer(argv[1]);

            // connect gem_pointer to my_pixBlock
            gem_state_pointer->image=&my_pixBlock;
            gem_state_pointer->texCoords=0;

            if (camera_started[_camID])   // if camera stopped : bypass all that stuf
            {
                // check if resolution has changed
                if (_resolution!=resolution[_camID])
                {
                    _resolution=resolution[_camID];
                    if (_resolution==CLEYE_QVGA)
                    {
                        _width=320;
                        _height=240;
                    }
                    else
                    {
                        _width=640;
                        _height=480;
                    }
                    clear_buffer(pBuffer, _channels*_width*_height);
                    my_pixBlock.image.xsize= _width;
                    my_pixBlock.image.ysize= _height;
                }
/// AJOUTER UPSIDE DOWN  FLIP MANUEL ICI
                // copy image from ps3eye driver
                CLEyeCameraGetFrame(pCam[_camID], pBuffer, 0);

                // check for colorspace conversion
                if (_colorspace==GL_RGBA)
                    BGRAtoRGBA(pBuffer,_width*_height);
       /*         else if (_colorspace==GL_RGB)
                    BGRtoRGB(pBuffer,_width*_height);
*/
                my_pixBlock.newimage=1;


            }
        }
    // forward gem_state pointers to outlet
    ToOutAnything(0,gensym("gem_state"),2,argv);
    }
}
        //dynamically changing framerate
void pix_ps3eye::set_frameRate(float f)
{
    float _frameRate=30;
    // check camera resolution

    // first case : QVGA (320x240)

    if (resolution[_camID]==CLEYE_QVGA)
    {
        // 0.1< f < 187
        _frameRate = max(0.1,min(187,f));
     }
    else
    {
        // 0.1< f < 75
        _frameRate = max(0.1,min(75,f));

    }

    //check if restarting cam is necessary
    if (_frameRate!=frameRate[_camID])
    {
        frameRate[_camID]=_frameRate;
        reset_cam();
    }
}
        //dynamically changing resolution

void pix_ps3eye::set_resolution (int argc,t_atom *argv)
{
    int _frameRate;
    CLEyeCameraResolution current_resolution;

    if (argc==1)
        if (IsSymbol(argv[0]))
        {
            current_resolution=resolution[_camID];
            const t_symbol *s = GetSymbol(argv[0]);
            if (s==resolution_QVGA)
                resolution[_camID]=CLEYE_QVGA;
            if (s==resolution_VGA)
                resolution[_camID]=CLEYE_VGA;
            // check if resolution changed
            if (resolution[_camID]!=current_resolution)
            {
                 _frameRate=frameRate[_camID];
                //force camera re-instantiation before  set_frameRate() call
                frameRate[_camID]=0;
                set_frameRate(_frameRate);

            }
        }

}

        //dynamically changing colorspace

void pix_ps3eye::set_colorspace(int argc,t_atom *argv)
{
    CLEyeCameraColorMode _colorMode;
    if(argc==1)
        if (IsSymbol(argv[0]))
        {
            const t_symbol *s = GetSymbol(argv[0]);
            // matching colorspace name
            if (s==color_RGBA)
            {
                my_pixBlock.image.csize=_channels=4;
                my_pixBlock.image.format=_colorspace=GL_RGBA;
                clear_buffer(pBuffer,_width*_height*_channels);
                if (enable_image_transform[_camID])
                    _colorMode=CLEYE_COLOR_PROCESSED;
                else _colorMode=CLEYE_COLOR_RAW;
            }
   /*         if (s==color_RGB)
            {
                my_pixBlock.image.csize=_channels=3;
                my_pixBlock.image.format=_colorspace=GL_RGB;
                clear_buffer(pBuffer,width[_camID]*height[_camID]*_channels);
            }
            */
             if (s==color_BGRA)
            {
                my_pixBlock.image.csize=_channels=4;
                my_pixBlock.image.format=_colorspace=GL_BGRA;
                clear_buffer(pBuffer,_width*_height*_channels);
                if (enable_image_transform[_camID])
                    _colorMode=CLEYE_COLOR_PROCESSED;
                else _colorMode=CLEYE_COLOR_RAW;
            }
  /*          if (s==color_BGR)
            {
                my_pixBlock.image.csize=_channels=3;
                my_pixBlock.image.format=_colorspace=GL_BGR;
                clear_buffer(pBuffer,width[_camID]*height[_camID]*_channels);
            }
            if (s==color_BGR16)
            {
                my_pixBlock.image.csize=_channels=2;
                my_pixBlock.image.format=_colorspace=GL_RGB16;
                clear_buffer(pBuffer,width[_camID]*height[_camID]*_channels);
            }
        */
            if ((s==color_Grey)||(s==color_GREY))
            {
                my_pixBlock.image.csize=_channels=1;
                my_pixBlock.image.format=_colorspace=GL_LUMINANCE;
                clear_buffer(pBuffer,_width*_height*_channels);
                if (enable_image_transform[_camID])
                    _colorMode=CLEYE_MONO_PROCESSED;
                else _colorMode=CLEYE_MONO_RAW;
            }

            if (_colorMode!=mode[_camID])
            // we need to restart / reset camera instance
            {
                mode[_camID]=_colorMode;
                reset_cam();
            }

        }

}

void pix_ps3eye::clear_buffer(PBYTE myBuffer, long size)
    {
        long index=size;
        while (index > 0)
        {
            myBuffer[--index] = 0;
        }
    }


void pix_ps3eye::BGRAtoRGBA(PBYTE myBuffer, int pixel_size)
    {
        BYTE mybyte;
        for (int i=0;i<(pixel_size*4);)
        {
                mybyte=myBuffer[i];
                myBuffer[i]=myBuffer[i+2];
                myBuffer[i+2]=mybyte;
                i+=4;
        }
    }
void pix_ps3eye::BGRtoRGB(PBYTE myBuffer, long pixel_size)
    {
        BYTE mybyte;
        for (int i=0;i<(pixel_size*3);)
        {
            mybyte=myBuffer[i];
            myBuffer[i]=myBuffer[i+2];
            myBuffer[i+2]=mybyte;
            i+=3;
        }
    }

void pix_ps3eye::set_gain(int i)
    {
        i=min(max(0,i),79);
        CLEyeSetCameraParameter(pCam[_camID],CLEYE_GAIN,i);
    }

void pix_ps3eye::set_AGC(int i)
    {
       CLEyeSetCameraParameter(pCam[_camID],CLEYE_AUTO_GAIN,(i==1));
    }

void pix_ps3eye::set_exposure(int i)
    {
        i=min(max(0,i),511);
        CLEyeSetCameraParameter(pCam[_camID],CLEYE_EXPOSURE,i);
    }

void pix_ps3eye::set_AEC(int i)
    {
        CLEyeSetCameraParameter(pCam[_camID],CLEYE_AUTO_EXPOSURE,(i==1));
    }
void pix_ps3eye::set_red(int i)
    {
        i=min(max(0,i),255);
        CLEyeSetCameraParameter(pCam[_camID],CLEYE_WHITEBALANCE_RED,i);
    }

void pix_ps3eye::set_green(int i)
    {
        i=min(max(0,i),255);
        CLEyeSetCameraParameter(pCam[_camID],CLEYE_WHITEBALANCE_GREEN,i);
    }

void pix_ps3eye::set_blue(int i)
    {
         i=min(max(0,i),255);
        CLEyeSetCameraParameter(pCam[_camID],CLEYE_WHITEBALANCE_BLUE,i);
    }

void pix_ps3eye::set_AWB(int i)
    {
        CLEyeSetCameraParameter(pCam[_camID],CLEYE_AUTO_WHITEBALANCE,(i==1));
    }

void pix_ps3eye::set_image_transform_mode(int i)
    {
      int _enable_image_transform=enable_image_transform[_camID];
        CLEyeCameraColorMode _color_mode=mode[_camID];
        if (i)
            switch (_color_mode)  //CLEYE_MONO_PROCESSED,CLEYE_COLOR_PROCESSED,CLEYE_MONO_RAW,CLEYE_COLOR_RAW,	CLEYE_BAYER_RAW
            {
                case CLEYE_MONO_RAW:
                {
                    _color_mode=CLEYE_MONO_PROCESSED;
                    _enable_image_transform=1;
                    break;
                }
                case CLEYE_COLOR_RAW:
                {
                    _color_mode=CLEYE_COLOR_PROCESSED;
                    _enable_image_transform=1;
                    break;
                }
                default:
                    break;
        }
        else
            switch (_color_mode)  //CLEYE_MONO_PROCESSED,CLEYE_COLOR_PROCESSED,CLEYE_MONO_RAW,CLEYE_COLOR_RAW,	CLEYE_BAYER_RAW
            {
                case CLEYE_MONO_PROCESSED:
                {
                    _color_mode=CLEYE_MONO_RAW;
                    _enable_image_transform=0;
                    break;
                }
                case CLEYE_COLOR_PROCESSED:
                {
                    _color_mode=CLEYE_COLOR_RAW;
                    _enable_image_transform=0;
                    break;
                }
                default:
                    break;
        }

        if (_color_mode != mode[_camID])
        {// we need to restart / reset camera instance
            mode[_camID]=_color_mode;
            enable_image_transform[_camID]=_enable_image_transform;
            reset_cam();
        }
     }


    void pix_ps3eye::set_hflip(int i)
    {
        CLEyeSetCameraParameter(pCam[_camID],CLEYE_HFLIP,(i==1));
    }
    void pix_ps3eye::set_vflip(int i)
    {
        CLEyeSetCameraParameter(pCam[_camID],CLEYE_VFLIP,(i==0));//top of image is inverted in GEM
    }
    void pix_ps3eye::set_hkeystone(int i)
    {
        i=min(max(-500,i),500);
        CLEyeSetCameraParameter(pCam[_camID],CLEYE_HKEYSTONE,i);
    }
    void pix_ps3eye::set_vkeystone(int i)
    {
        i=min(max(-500,i),500);
        CLEyeSetCameraParameter(pCam[_camID],CLEYE_VKEYSTONE,i);
    }
    void pix_ps3eye::set_xoffset(int i)
    {
        i=min(max(-500,i),500);
        CLEyeSetCameraParameter(pCam[_camID],CLEYE_XOFFSET,i);
    }
    void pix_ps3eye::set_yoffset(int i)
    {
        i=min(max(-500,i),500);
        CLEyeSetCameraParameter(pCam[_camID],CLEYE_YOFFSET,i);
    }
    void pix_ps3eye::set_rotation(int i)
    {
        i=min(max(-500,i),500);
        CLEyeSetCameraParameter(pCam[_camID],CLEYE_ROTATION,i);
    }
    void pix_ps3eye::set_zoom(int i)
    {
        i=min(max(-500,i),500);
        CLEyeSetCameraParameter(pCam[_camID],CLEYE_ZOOM,i);
    }
    	// camera non-linear transform parameters
    void pix_ps3eye::set_lenscorrection1(int i)
    {
        i=min(max(-500,i),500);
        CLEyeSetCameraParameter(pCam[_camID],CLEYE_LENSCORRECTION1,i);
    }
    void pix_ps3eye::set_lenscorrection2(int i)
    {
        i=min(max(-500,i),500);
        CLEyeSetCameraParameter(pCam[_camID],CLEYE_LENSCORRECTION2,i);
    }
    void pix_ps3eye::set_lenscorrection3(int i)
    {
        i=min(max(-500,i),500);
        CLEyeSetCameraParameter(pCam[_camID],CLEYE_LENSCORRECTION3,i);
    }
    void pix_ps3eye::set_lensbrightness(int i)
    {
        i=min(max(-500,i),500);
        CLEyeSetCameraParameter(pCam[_camID],CLEYE_LENSBRIGHTNESS,i);
    }

   void pix_ps3eye::set_upsidedown(int i)
        {
                my_pixBlock.image.upsidedown = (i==1);
        }

    // method for manually start / stop cam capture
    void pix_ps3eye::start_cam(int i)
    {
        if (!i)
        {
            if (camera_started[_camID])
                camera_started[_camID]=(!CLEyeCameraStop(pCam[_camID]));
        }
        else
            if (!camera_started[_camID])
            {
                camera_started[_camID]=CLEyeCameraStart(pCam[_camID]);
            }
    }
    void pix_ps3eye::reset_cam()
    {
        int _vflip=CLEyeGetCameraParameter(pCam[_camID], CLEYE_VFLIP);
        if (camera_started[_camID])
        CLEyeCameraStop(pCam[_camID]);

        // destroy instance and recreate

        CLEyeDestroyCamera(pCam[_camID]);
        pCam[_camID]=NULL;
        pCam[_camID] = CLEyeCreateCamera(cameraGUID[_camID], mode[_camID], resolution[_camID], frameRate[_camID]);


        if (pCam[_camID])
        {
            CLEyeSetCameraParameter(pCam[_camID], CLEYE_VFLIP,_vflip);
            if (camera_started[_camID])     CLEyeCameraStart(pCam[_camID]);
        }
        else post ("error resetting camera %d", _camID+1);
    }
