#ifndef PIX_PS3EYE_H_INCLUDED
#define PIX_PS3EYE_H_INCLUDED

// include flext header
#include <flext.h>

// check for appropriate flext version
#if !defined(FLEXT_VERSION) || (FLEXT_VERSION < 500)
#error You need at least flext version 0.5.0
#endif

//  header for GEM compatibility
#include "gem_headers.h"

//  header for Ps3Eye SDK
#include "CLEyeMulticam.h"

class pix_ps3eye:
	public flext_base
{
	FLEXT_HEADER_S(pix_ps3eye,flext_base,setup)

public:
	// constructor
	pix_ps3eye(int argc,t_atom *argv);
	//destructor
	~pix_ps3eye();

    // table registering how many pix_ps3eye objects are created for each ps3Eye cam
    static int instanciated[];
    static bool camera_started[];

    // tables of PS3EyeLib objects
    static CLEyeCameraInstance pCam[];
    static GUID cameraGUID[];

 // tables registering capture setting for each ps3eye cam

	static CLEyeCameraColorMode mode[];
	static CLEyeCameraResolution resolution[];
    static float frameRate[];
    static bool enable_image_transform[];

protected:

    void gem_state(int argc, t_atom *argv);
    void set_frameRate(float f);
    void set_resolution(int argc,t_atom *argv);
    void set_colorspace(int argc,t_atom *argv);
    void clear_buffer(PBYTE myBuffer, long size);
        // colorspace conversion functions
    void BGRAtoRGBA(PBYTE myBuffer, int pixel_size);
    void BGRtoRGB(PBYTE myBuffer, long pixel_size);

    void set_gain(int i);
    void set_AGC(int i);
    void set_exposure(int i);
    void set_AEC(int i);
    void set_red(int i);
    void set_green(int i);
    void set_blue(int i);
    void set_AWB(int i);
           // camera linear transform parameters
    void set_image_transform_mode(int i);

    void set_hflip(int i);
    void set_vflip(int i);
    void set_hkeystone(int i);
    void set_vkeystone(int i);
    void set_xoffset(int i);
    void set_yoffset(int i);
    void set_rotation(int i);
    void set_zoom(int i);
    	// camera non-linear transform parameters
    void set_lenscorrection1(int i);
    void set_lenscorrection2(int i);
    void set_lenscorrection3(int i);
    void set_lensbrightness(int i);
   void set_upsidedown(int i);

    // method for manually start / stop cam capture
    void start_cam(int i);
    void reset_cam();

private:
    float arg;  // argument variable

    // pix_ps3eye instance-specific image transfert parameters
    int _channels; // 1 (gray)  / 3 (BGR)   / 4  (BGRA)....

    int format_index;  /// obsolete

    unsigned int _colorspace;  // GL colorspace naming convention


    CLEyeCameraResolution _resolution;
    GUID _cameraGUID;

    BYTE r,g,b;
    int _width;
    int _height;

    // index specifying which cam is used by this instance of pix_ps3eye
    int _camID;

    // image origin(top-left or bottom-left)
    /// obsolete
    int flip;

//  entries in table of symbols for colorspace names
    const t_symbol *color_RGBA, *color_RGB, *color_BGRA, *color_BGR, *color_BGR16,
    *color_Grey, *color_GREY, *resolution_QVGA, *resolution_VGA;


    // image buffer and pix structures
    pixBlock my_pixBlock;
    imageStruct my_imageStruct;
    PBYTE pBuffer;
    GemState *gem_state_pointer;

    t_atom my_output;

    FLEXT_CALLBACK_G(gem_state)
    FLEXT_CALLBACK_F(set_frameRate)
    FLEXT_CALLBACK_V(set_resolution)

    FLEXT_CALLBACK_I(set_gain)
    FLEXT_CALLBACK_I(set_AGC)
    FLEXT_CALLBACK_I(set_exposure)
    FLEXT_CALLBACK_I(set_AEC)
    FLEXT_CALLBACK_I(set_red)
    FLEXT_CALLBACK_I(set_green)
    FLEXT_CALLBACK_I(set_blue)
    FLEXT_CALLBACK_I(set_AWB)
    FLEXT_CALLBACK_V(set_colorspace)
    FLEXT_CALLBACK_I(set_upsidedown)
    FLEXT_CALLBACK_I(start_cam)
    FLEXT_CALLBACK(reset_cam)
    FLEXT_CALLBACK_I(set_image_transform_mode)
    FLEXT_CALLBACK_I(set_hflip)
    FLEXT_CALLBACK_I(set_vflip)
    FLEXT_CALLBACK_I(set_hkeystone)
    FLEXT_CALLBACK_I(set_vkeystone)
    FLEXT_CALLBACK_I(set_xoffset)
    FLEXT_CALLBACK_I(set_yoffset)
    FLEXT_CALLBACK_I(set_rotation)
    FLEXT_CALLBACK_I(set_zoom)
    FLEXT_CALLBACK_I(set_lenscorrection1)
    FLEXT_CALLBACK_I(set_lenscorrection2)
    FLEXT_CALLBACK_I(set_lenscorrection3)
    FLEXT_CALLBACK_I(set_lensbrightness)


//setup function : initialise static tables
	static void setup(t_classid c);


};


#endif // PIX_PS3EYE_H_INCLUDED
